# 
# Exodb
# Copyright (C) 2014
#
# author: Natapol Pornputtapong <natapol.por@gmail.com>
#
# Documentation: Natapol Pornputtapong (RDoc'd and embellished by William Webber)
#

# raise "Please, use ruby 1.9.0 or later." if RUBY_VERSION < "1.9.0"

module Exodb
    
	module GeneLocationField
		
		extend ActiveSupport::Concern
		
		included do
			field :start,				type: Integer
			field :stop,				type: Integer
			field :chr,					type: String
			field :seqstart,			type: Integer
			field :seqstop,				type: Integer
			field :strand,				type: String
			field :assembly,			type: String
			field :sequence,			type: String
			field :chrrefseq,			type: String # refseq id of chromosome
			
			validates_presence_of :start, message: "start field missing"
			validates_presence_of :stop, message: "stop field missing"
			validates_presence_of :chr, message: "chr field missing"
			validates_presence_of :assembly, message: "assembly field missing"
			
			index({start: 1, stop: 1, chr: 1, assembly: 1}, background: true)
		end
		
		module ClassMethods
			
			def where_cover(loc_str)
				
				dat = parse_locstr(loc_str)
				
				return self.where({chr: dat['chr'], assembly: dat['assembly']}).lte(start: dat['start']).gte(stop: dat['stop'])
				
			end
			
			def where_intersect(loc_str)
				dat = parse_locstr(loc_str)
				
				return self.where({chr: dat['chr'], assembly: dat['assembly']}).or({:start.lte => dat['start'], :stop.gte => dat['start']}, {:start.lte => dat['stop'], :stop.gte => dat['stop']})
			end
			
			def where_in(loc_str)
				dat = parse_locstr(loc_str)
				return self.where({chr: dat['chr'], assembly: dat['assembly']}).gte(start: dat['start']).lte(stop: dat['stop'])
			end
			
			def where_ups_cover(loc_str)
				#code
			end
			
		end
		
		# Download sequence from web service please use by caution. NCBI will block scamming sequest
		#
		def dl_seq!
			
			case self.chrrefseq
			when /\Aurn:miriam:refseq:/
				self.sequence = Bio::FastaFormat.new(Bio::NCBI::REST.efetch(self.chrrefseq.split(':', 4), {"db"=>"nucleotide", "rettype"=>"fasta", "retmode"=>"text", "seq_start"=>self.start, "seq_stop"=>self.end})).seq
			else 
				self.sequence = Exodb::Ensembl::REST.sequence_region()
			end
			
			self.save!
			
		end
		
		# get the start position of gene rely on the genome
		#
		# @return [Integer] start position of gene
		def start
			self[:start]
		end
		
		# get the end position of gene rely on the genome
		#
		# @return [Integer] end position of gene
		def stop
			self[:stop]
		end
		
		# get the start position of gene rely on the genome
		#
		# @return [Integer] start position of gene
		def begin
			self[:strand] == '+' ? self[:start] : self[:stop]
		end
		
		# get the start position of gene rely on the genome
		#
		# @return [Integer] start position of gene
		def end
			self[:strand] == '+' ? self[:stop] : self[:start]
		end
		
		# get the chromosome
		#
		# @return [Integer] chromosome
		def chromosome
			self[:chr]
		end
		
		# Assign location
		#
		# @param [String, Hash] location string in chromosome:start..stop or chromosome:start-stop format
		def location=(loc)
			if loc.is_a?(String)
				
				begin
					loc.parse_loc.delete_if {|k, v| k == 'pos'}.each_pair do |k, v|
						self[k.to_sym] = v
					end
				rescue
					
				end
				
			end
		end
		
		# Return location
		#
		# @return [String] location string in chromosome:position
		def location_str
			return "#{self.chromosome}:#{[self.start, self.stop].uniq.join('..')}"
		end
		
		alias_method :locstr, :location_str
		
		# Return gene sequence
		#
		# @return [Bio::Sequence] gene sequence
		def to_seq
			whole_seq.splice("#{self[:start] - self[:seqstart] + 1}..#{self[:stop] - self[:seqstart] + 1}")
		end
		
		# Return whole deposited sequence
		#
		# @return [Bio::Sequence] gene sequence
		def whole_seq
			Bio::Sequence::NA.new(self[:sequence])
		end
		
		# join exon or cds position into a string
		#
		# @param [Array] input array exon or cds
		# @param [Interger] Position to stop positive value for forward read negative value for complement
		#
		# @return [String] a string in start..end,start..end,...
		def get_splice(arr, strand = nil)
			
			strand = strand || self[:strand]
			
			reducer = self[:seqlocation]['start'] - 1
			
			str = []
			
			arr.each do |e|
				str.push("#{e[0] - reducer}..#{e[1] - reducer}")
			end
			
			return strand == '+' ? self.to_seq.splicing("join(#{str.join(',')})") : self.to_seq.splicing("complement(join(#{str.join(',')}))")
			
		end
	end
    
end