# 
# Exodb
# Copyright (C) 2014
#
# author: Natapol Pornputtapong <natapol.por@gmail.com>
#
# Documentation: Natapol Pornputtapong (RDoc'd and embellished by William Webber)
#

# raise "Please, use ruby 1.9.0 or later." if RUBY_VERSION < "1.9.0"

module Exodb
	
	class Chrref < Reference
		
		include Exodb::GeneLocationField
		
		index({oid: 1}, background: true)
		
	end
	
	class Generef < Reference
		
		include Exodb::GeneLocationField
		
		@@expanding = 500
		
		field :psuedo,				type: Boolean
		
		index({sequence: 'text'}, background: true)
		index({oid: 1, psudo: 1, chrrefseq: 1, assembly: 1}, background: true)
		
		has_many :genes
		embeds_many :isoforms
		embeds_many :occurrents
		
		validates_format_of :chrrefseq, with: /\A(urn:miriam:refseq)/
		
		#oid = "chrrefseq:start..stop"
		
		def self.expanding
			return @@expanding
		end
		
		# Download gene symbol from HGNC service
		#
		def dl_symbol!
			
			baseuri = "http://rest.genenames.org/search"
			
			
			query = ""
			
			if self.get_xref('urn:miriam:refseq')
				query = "#{baseuri}/refseq_accession/#{self.chrrefseq.id.split('.')[0]}"
			elsif self.get_xref('urn:miriam:ncbigene')
				query = ""
			end
			
			if !query.empty?
				response = JSON.parse(open(query, 'Accept' => 'application/json').read)['response']
				if !response['docs'].empty?
					
					response['docs'].each do |e|
						self.add_to_set(:xrefs, "urn:miriam:hgnc:#{e["hgnc_id"]}")
						self.add_to_set(:xrefs, "urn:miriam:hgnc.symbol:#{e["symbol"]}")
					end
					
					self.save!
					
				end
				
			end
			
		end
		
		# Download incident data from TCGA
		#
		def dl_occurrent!
			
			if self.get_xref('urn:miriam:hgnc.symbol')
				
				cancerstudies = []
				
				open("http://www.cbioportal.org/public-portal/webservice.do?cmd=getCancerStudies") {|f|
					f.each_line {|line| cancerstudies.push(line.chomp.split("\t")[0])}
				}
				
				occurrents = {}
				totalcase = {}
				
				cancerstudies.each do |study|
					
					totalcase[study] = 0 if !totalcase.has_key?(study)
					
					open("http://www.cbioportal.org/public-portal/webservice.do?cmd=getCaseLists&cancer_study_id=#{study}") do |f|
						f.each_line do |line|
							totalcase[study] += line.chomp.split(/\t/)[4].split(' ').length if line =~ /\tSequenced Tumors\t/
						end
					end
					
					occurrents[study] = {} if !occurrents.has_key?(study)
					
					open("http://www.cbioportal.org/public-portal/webservice.do?cmd=getMutationData&genetic_profile_id=#{study}_mutations&gene_list=#{self.get_xref('urn:miriam:hgnc.symbol').id}") do |f|
						f.each_line do |line|
							dat = line.chomp.split(/\t/)
							
							if dat[5] == 'Missense_Mutation'
								
								occurrents[study][dat[7].split(/(\d+)/)[1]] = [] if !occurrents[study].has_key?(dat[7].split(/(\d+)/)[1])
								occurrents[study][dat[7].split(/(\d+)/)[1]].push(dat[2])
								
							end
							
						end
						
						
					end
				end
				
				self.occurrents.clear if self.occurrents
				
				occurrents.each_pair do |cancertype, v|
					v.each_pair do |position, occur|
						self.occurrents << Occurrent.new({cancertype: cancertype, position: position, occur: occur.uniq.sort, casenumber: totalcase[cancertype]})
					end
				end
				
				self.save!
				
			end
			
		end
		
		# return longest splice of this gene
		def longest_splice()
			length = 0
			longest = nil
			self.isoforms.each do |e|
				
				if e.prot_len > length
					length = e.prot_len
					longest = e
				end
				
			end
			
			return longest
		end
		
		alias_method :longest_isoform, :longest_splice
		
		# Check that this gene has any splice variant
		#
		# @return [Boolean] true if has any splices
		def has_splices?
			return self.isoforms.exists?
		end
		
		alias_method :has_isoforms?, :has_splices?
		
		# Check if Generef has sequence
		#
		# @return [Boolean] Return true if there is a sequence
		def has_sequence?()
			return self[:sequence] ? true : false
		end
		
		# Check if Generef can translate
		#
		# @return [Boolean] Return true if this can be translate
		def can_translated?()
			return self.has_sequence? && self.has_splices? && self.longest_splice != nil ? true : false
		end
		
		# Get gene symbol
		#
		# @return [String] Return gene symbol or any id from xrefs or 'nosymbol'
		def symbol
			
			if self.get_xref('urn:miriam:hgnc.symbol')
				return self.get_xref('urn:miriam:hgnc.symbol').id
			elsif self.xrefs && !self.xrefs.empty?
				return self.xrefs.sort[0].id
			else
				return 'nosymbol'
			end
			
		end
		
	end
	
end